var syntaxError = require('syntax-error');
var fs = require('fs');
var chokidar = require('chokidar');
var eventEmitter = new (require('events').EventEmitter);

var jobList = {};
var watchers = [];

function JobError(message, name) {
  Error.captureStackTrace(this, JobError);
  this.name = name+'Error';
  this.message = message;
  return this;
}

var onReload = function() {}
var requireError = function(e, jobName, module, reject) {
  if(e instanceof SyntaxError) {
    var resolved = require.resolve(module);
    fs.readFile(resolved, 'utf8', function(err, content) {
      console.log(syntaxError(content,resolved));
      reject(new JobError('Failed to register job '+jobName,'Require'))
    })
  } else {
    if(e.code === 'MODULE_NOT_FOUND')
    console.log(e.toString());
    reject(new JobError('Failed to register job '+jobName,'Require'))
  }
}

var jobs = {
  register: function(jobName, module, watch) {
    return new Promise(function(resolve, reject) {
      if(watch === undefined) watch = true;
      if(jobList[jobName] !== undefined) {
        return Promise.reject(
          new JobError('Job '+jobName+' is already registered', 'JobAlreadyRegistered')
        )
      }

      var f;
      try {
        f = require(module);
      } catch(e) {
        requireError(e, jobName, module, reject);
        return;
      }

      if(typeof f !== 'function')
        throw new JobError('Module "'+module+'" for job "'+jobName+'" does not export function');
      jobList[jobName] = {
        f: f,
        module: module
      };
      if(watch) jobs.watch(jobName);
      resolve();

    });
  },
  reload: function(jobName) {
    var module = jobList[jobName].module;
    delete require.cache[require.resolve(module)];

    return new Promise(function(resolve, reject) {
      try {
        var f = require(module);
        if(typeof f === 'function') {
          var job = {
            f: f,
            module: module
          };
          delete jobList[jobName];
          jobList[jobName] = job;
          resolve();
        } else {
          console.log('Error reloading job '+jobName+' retry in 500ms');
          setTimeout(function() {
            resolve(jobs.reload(jobName))
          },500)
        }
      } catch(e) {
        console.log(e);
        requireError(e, jobName, module, reject);
      }
    });
  },
  // registerMultiple(
  //  {watch: true},
  //  ['jobName','./module'],['jobName2','./module2']
  // )
  registerMultiple: function() {
    var regs = [];
    var watch = true;
    for(i in arguments) {
      var job = arguments[i];
      if(i == 0) {
        if(job.watch !== undefined) watch = job.watch;
      } else {
        regs.push(jobs.register(job[0], job[1], watch));
      }
    }
    return Promise.all(regs);
  },
  run: function() {
    var jobName = arguments[0];
    if(jobList[jobName] === undefined) {
      throw new JobError('Job '+jobName+' is not registered', 'JobNotRegistered')
    }
    try {
      return jobList[jobName].f.apply(null, Array.prototype.slice.call(arguments, 1));
    } catch(e) {
      console.log("Error running job",jobName);
      console.log(jobList[jobName]);
      throw e;
    }
  },
  runSequence: function() {
    var prom;
    console.log("Running",Array.prototype.slice.call(arguments, 0))
    var jobSequence = [].splice.call(arguments, 0);
    jobSequence.forEach(function(job){
      if(prom === undefined) {
        prom = jobs.run(job);
      } else {
        prom = prom.then(function() {
          var args = [].splice.call(arguments, 0);
          args.splice(0,0,job);
          return jobs.run.apply(null,args)
        })
      }
    });
    return prom;
  },
  watch: function(jobName) {
    var module = jobList[jobName].module;
    var file = require.resolve(module);
    var w = chokidar.watch(file);
    w.on('change', path => {
      jobs.reload(jobName)
      .then(onReload);
    });
    watchers.push(w);
  },
  onReload: function(f) {
    onReload = f;
  },
  close: function() {
    watchers.forEach(function(w) {
      w.close();
      console.log(w.getWatched());
    });
    watchers = [];
  }
}

module.exports = jobs;
